unit Unit1;

{$mode objfpc}{$H+}

interface

uses
  Classes, SysUtils, Forms, Controls, Graphics, Dialogs, laz.VirtualTrees;

type

  { TForm1 }

  TForm1 = class(TForm)
    Tree: TLazVirtualDrawTree;
    procedure FormCreate(Sender: TObject);
    procedure FormDestroy(Sender: TObject);
    procedure TreeDrawNode(Sender: TBaseVirtualTree;
      const PaintInfo: TVTPaintInfo);
    procedure TreeGetNodeWidth(Sender: TBaseVirtualTree; HintCanvas: TCanvas;
      Node: PVirtualNode; Column: TColumnIndex; var NodeWidth: Integer);
    procedure TreeMeasureItem(Sender: TBaseVirtualTree; TargetCanvas: TCanvas;
      Node: PVirtualNode; var NodeHeight: Integer);
  private
    function GetBild(Sender: TBaseVirtualTree; Node: PVirtualNode): TCustomBitmap;
    function GetCellText(Sender: TBaseVirtualTree; Node: PVirtualNode): String;
    procedure PopulateTree;

  public

  end;

var
  Form1: TForm1;

type
  TAdressbuchRec = record
    Name: String;
    Anschrift: String;
    Telefon: String;
    Geburtstag: TDate;
    www: String;
    Bild: TJpegImage;
  end;

var
  Adressbuch: Array of TAdressbuchRec;

implementation

{$R *.lfm}

type
  TAdressTyp = (atName, atAnschrift, atTelefon, atGeburtstag, atWWW, atBild, atValue);

  PAdressNodeData = ^TAdressNodeData;
  TAdressNodeData = record
    Index: Integer;   // Index im Adressbuch-Array
    Typ: TAdressTyp;
  end;

const
  BILD_HOEHE = 120;
  TEXT_MARGIN = 2;

{ TForm1 }

procedure TForm1.FormCreate(Sender: TObject);
var
  pic: TPicture;
  R: TRect;
begin
  SetLength(Adressbuch, 1);

  Adressbuch[0].Name := 'Diehaare Liegen';
  Adressbuch[0].Anschrift := 'Auf der Straße 7';
  Adressbuch[0].Telefon := '0123 456 789';
  Adressbuch[0].Geburtstag := EncodeDate(1977, 7, 7);
  Adressbuch[0].www := 'https://diehaareliegen.de';

  pic := TPicture.Create;
  try
    pic.LoadfromFile('paw.png');
    R := Rect(0, 0, round(pic.Width/pic.Height*BILD_HOEHE), BILD_HOEHE);
    Adressbuch[0].Bild := TJpegImage.Create;
    Adressbuch[0].Bild.SetSize(R.Width, R.Height);
    Adressbuch[0].Bild.Canvas.StretchDraw(R, pic.Bitmap);
  finally
    pic.Free;
  end;

  Tree.NodeDataSize := SizeOf(TAdressNodeData);
  Tree.TreeOptions.MiscOptions := Tree.TreeOptions.MiscOptions + [toVariableNodeHeight];
  PopulateTree;
end;

procedure TForm1.FormDestroy(Sender: TObject);
var
  i: Integer;
begin
  for i := 0 to High(Adressbuch) do
    Adressbuch[i].Bild.Free;
end;

{ Ermittelt das Bild, falls Node der Bild-Node ist. }
function TForm1.GetBild(Sender: TBaseVirtualTree; Node: PVirtualNode): TCustomBitmap;
var
  data, parentData: PAdressNodeData;
begin
  data := Sender.GetNodeData(Node);
  parentData := Sender.GetNodedata(Node^.Parent);
  if (parentData <> nil) and (parentData^.Typ = atBild) then
    Result := Adressbuch[data^.Index].Bild
  else
    Result := nil;
end;

{ Ermittelt den Text, der in der Zelle stehen soll. }
function TForm1.GetCellText(Sender: TBaseVirtualTree; Node: PVirtualNode): String;
const
  NODE_NAME: array[TAdressTyp] of string = ('', '', 'Telefon', 'Geburtstag', 'WWW', 'Bild', '');
var
  data: PAdressNodeData;
  parentData: PAdressNodeData;
  idx: Integer;
begin
  Result := '';
  data := Sender.GetNodeData(Node);
  if Node^.Parent <> nil then
    parentData := Sender.GetNodeData(Node^.Parent)
  else
    parentData := nil;
  idx := data^.Index;
  case data^.Typ of
    atName:
      Result := Adressbuch[idx].Name;
    atAnschrift:
      Result := Adressbuch[idx].Anschrift;
    atTelefon, atGeburtstag, atWWW, atBild:
      Result := NODE_NAME[data^.Typ];
    atValue:
      case parentdata^.Typ of
        atName: ;       // schon behandelt
        atAnschrift: ;  // dto.
        atTelefon: Result := Adressbuch[idx].Telefon;
        atGeburtstag: Result := DateToStr(Adressbuch[idx].Geburtstag);
        atWWW: Result := Adressbuch[idx].WWW;
        atBild: if Adressbuch[idx].Bild = nil then Result := '(ohne Bild)' else Result := '';
      end;
  end;
end;


{ Zeichnet den Node, primär den Text, aber beim BildNode auch das Bild. }
procedure TForm1.TreeDrawNode(Sender: TBaseVirtualTree;
  const PaintInfo: TVTPaintInfo);
var
  s: String;
  img: TCustomBitmap;
  x, y: Integer;
begin
  if (PaintInfo.Column = Sender.FocusedColumn) and (PaintInfo.Node = Sender.FocusedNode) then
    PaintInfo.Canvas.Font.Color := clHighlightText
  else
    PaintInfo.Canvas.Font.Color := clWindowText;

  x := PaintInfo.ContentRect.Left + TEXT_MARGIN;
  y := PaintInfo.ContentRect.Top + TEXT_MARGIN;

  img := GetBild(Sender, PaintInfo.Node);
  if img <> nil then
    PaintInfo.Canvas.Draw(x, y, img)
  else
  begin
    s := GetCellText(Sender, PaintInfo.Node);
    PaintInfo.Canvas.Brush.Style := bsClear;
    PaintInfo.Canvas.TextOut(x, y, s);
  end;
end;

{ Berechnet die Höhe des angegebenen Nodes aus Texthöhe bzw. Bildhöhe }
procedure TForm1.TreeGetNodeWidth(Sender: TBaseVirtualTree;
  HintCanvas: TCanvas; Node: PVirtualNode; Column: TColumnIndex;
  var NodeWidth: Integer);
var
  img: TCustomBitmap;
  s: String;
begin
  img := GetBild(Sender, Node);
  if img <> nil then
    NodeWidth := img.Width + 2*TEXT_MARGIN
  else
  begin
    s := GetCellText(Sender, Node);
    NodeWidth := Sender.Canvas.TextWidth(s) + 2*TEXT_MARGIN;
  end;
end;

{ Vergrößert die Zeile mit dem Bild. Wird wegen der gesetzten MiscOption
  toVariableNodeHeight aufgerufen. }
procedure TForm1.TreeMeasureItem(Sender: TBaseVirtualTree;
  TargetCanvas: TCanvas; Node: PVirtualNode; var NodeHeight: Integer);
var
  img: TCustomBitmap;
  s: String;
begin
  img := GetBild(Sender, Node);
  if img <> nil then
    NodeHeight := img.Height + 2 * TEXT_MARGIN
  else
  begin
    s := GetCellText(Sender, Node);
    NodeHeight := Sender.Canvas.TextHeight(s) + 2 * TEXT_MARGIN;
  end;
end;

procedure TForm1.PopulateTree;
var
  i: Integer;
  node, childNode, valueNode: PVirtualNode;
  data: PAdressNodeData;
begin
  for i := 0 to High(Adressbuch) do
  begin
    node := Tree.AddChild(nil);
    data := Tree.GetNodeData(node);
    data^.Index := i;
    data^.Typ := atName;

    childNode := Tree.AddChild(node);
    data := Tree.GetNodeData(childNode);
    data^.Index := i;
    data^.Typ := atAnschrift;

    childNode := Tree.AddChild(node);
    data := Tree.GetNodeData(childNode);
    data^.Index := i;
    data^.Typ := atTelefon;

    valueNode := Tree.AddChild(childnode);
    data := Tree.GetNodeData(valueNode);
    data^.Index := i;
    data^.Typ := atValue;

    childNode := Tree.AddChild(node);
    data := Tree.GetNodeData(childNode);
    data^.Index := i;
    data^.Typ := atGeburtstag;

    valueNode := Tree.AddChild(childnode);
    data := Tree.GetNodeData(valueNode);
    data^.Index := i;
    data^.Typ := atValue;

    childNode := Tree.AddChild(node);
    data := Tree.GetNodeData(childNode);
    data^.Index := i;
    data^.Typ := atWWW;

    valueNode := Tree.AddChild(childnode);
    data := Tree.GetNodeData(valueNode);
    data^.Index := i;
    data^.Typ := atValue;

    childNode := Tree.AddChild(node);
    data := Tree.GetNodeData(childNode);
    data^.Index := i;
    data^.Typ := atBild;

    valueNode := Tree.AddChild(childnode);
    data := Tree.GetNodeData(valueNode);
    data^.Index := i;
    data^.Typ := atValue;

    Tree.FullExpand(node);
  end;
end;

end.

